@@ -66,7 +66,7 @@ char user[USERLEN + 1];
 struct sockaddr_in sa;
 char pasv = 0;
 int sock;
-int pasvsock;
+int pasvsock = -1;
 char *philename = NULL;
 unsigned long offset = 0;
 short int xfertype = TYPE_BINARY;
@@ -136,7 +136,7 @@ int dataconn()
 	struct sockaddr foo;
 	struct sockaddr_in local;
 	socklen_t namelen = sizeof(foo);
-    int curuid = geteuid();
+        int curuid = geteuid();
 
 	memset(&foo, 0, sizeof(foo));
 	memset(&local, 0, sizeof(local));
@@ -156,7 +156,11 @@ int dataconn()
 		local.sin_addr.s_addr = name.sin_addr.s_addr;
 		local.sin_family = AF_INET;
         if (!strcasecmp(config_getoption("DATAPORT20"), "yes")) {
-            seteuid(0);
+            if (seteuid(0) < 0) {
+                control_printf(SL_FAILURE, "425-Fail to seteuid.\r\n425 %s.",
+                                strerror(errno));
+                return 1;
+            }
             local.sin_port = htons(20);
         }
 		if (bind(sock, (struct sockaddr *) &local, sizeof(local)) < 0) {
@@ -165,7 +169,11 @@ int dataconn()
 			return 1;
 		}
         if (!strcasecmp(config_getoption("DATAPORT20"), "yes"))
-            seteuid(curuid);
+            if (seteuid(curuid) < 0) {
+                control_printf(SL_FAILURE, "425-Fail to seteuid.\r\n425 %s.",
+                                strerror(errno));
+                return 1;
+            }
 		sa.sin_family = AF_INET;
 		if (connect(sock, (struct sockaddr *) &sa, sizeof(sa)) == -1) {
 			control_printf(SL_FAILURE, "425-Unable to establish data connection.\r\n"
@@ -213,12 +221,11 @@ void command_user(char *username)
 #ifdef DEBUG
 	bftpd_log("Trying to log in as %s.\n", user);
 #endif
-    expand_groups();
+        expand_groups();
 	if (!strcasecmp(config_getoption("ANONYMOUS_USER"), "yes"))
         {
                 state = STATE_USER;
                 control_printf(SL_SUCCESS, "331 Password please.");
-		/* bftpd_login(""); */
         }
 	else {
 		state = STATE_USER;
@@ -234,7 +241,10 @@ void command_pass(char *password)
 	}
 	if (bftpd_login(password)) {
 		bftpd_log("Login as user '%s' failed.\n", user);
-		control_printf(SL_FAILURE, "530 Login incorrect.");
+		/*
+                This is printed for us in the login function.
+                control_printf(SL_FAILURE, "530 Login incorrect.");
+                */
 		// exit(0);
                 state = STATE_CONNECTED;
                 return; 
@@ -262,8 +272,10 @@ void command_type(char *params)
     } else if ((*params == 'I') || (*params == 'i')) {
       	control_printf(SL_SUCCESS, "200 Transfer type changed to BINARY");
         xfertype = TYPE_BINARY;
-    } else
+    } else if (*params) {
         control_printf(SL_FAILURE, "500 Type '%c' not supported.", *params);
+    } else
+        control_printf(SL_FAILURE, "500 Type needs a parameter.");
 }
 
 void command_port(char *params) {
@@ -336,6 +348,9 @@ void command_pasv(char *foo)
         control_printf(SL_FAILURE, "500 EPSV ALL has been called.");
         return;
     }
+        /* avoid overflow by closing any previously existing passive sockets */
+        if (pasvsock >= 0)
+          close(pasvsock);
 	pasvsock = socket(AF_INET, SOCK_STREAM, 0);
 	sa.sin_addr.s_addr = INADDR_ANY;
 	sa.sin_family = AF_INET;
@@ -369,9 +384,9 @@ void command_pasv(char *foo)
             control_printf(SL_FAILURE, "425 Error: Unable to bind data socket.");
             return;
         }
-        prepare_sock(pasvsock);
     }    /* end of else using list of ports */
       
+        prepare_sock(pasvsock);
 	if (listen(pasvsock, 1)) {
 		control_printf(SL_FAILURE, "425-Error: Unable to make socket listen.\r\n425 %s",
 				 strerror(errno));
@@ -387,7 +402,7 @@ void command_pasv(char *foo)
             sscanf( my_override_ip, "%i.%i.%i.%i",
                     &a1, &a2, &a3, &a4);
         }
-        else   /* noraml, no over-ride */
+        else   /* normal, no over-ride */
         {
   	    sscanf((char *) inet_ntoa(name.sin_addr), "%i.%i.%i.%i",
 		   &a1, &a2, &a3, &a4);
@@ -418,15 +433,47 @@ void command_epsv(char *params)
             }
         }
     }
+    /* avoid overflow from multiple passive sockets */
+    if (pasvsock >= 0)
+       close(pasvsock);
     pasvsock = socket(AF_INET, SOCK_STREAM, 0);
     sa.sin_addr.s_addr = INADDR_ANY;
-    sa.sin_port = 0;
     sa.sin_family = AF_INET;
-    if (bind(pasvsock, (struct sockaddr *) &sa, sizeof(sa)) == -1) {
-		control_printf(SL_FAILURE, "500-Error: Unable to bind data socket.\r\n425 %s",
-				 strerror(errno));
-		return;
-	}
+    if (!config_getoption("PASSIVE_PORTS") || !strlen(config_getoption("PASSIVE_PORTS")))
+    { 
+        sa.sin_port = 0;
+        if (bind(pasvsock, (struct sockaddr *) &sa, sizeof(sa)) == -1) {
+	    control_printf(SL_FAILURE, "500-Error: Unable to bind data socket.\r\n425 %s",
+		           strerror(errno));
+	   return;
+        }   /* end of trying to bind */
+    }   /* end of there is no specific port range we need to use */
+
+
+    /* try to bind to a specific port range */
+    else {
+        int i = 0, success = 0, port;
+        for (;;) {
+            port = int_from_list(config_getoption("PASSIVE_PORTS"), i++);
+            if (port < 0)
+                break;
+            sa.sin_port = htons(port);
+            if (bind(pasvsock, (struct sockaddr *) &sa, sizeof(sa)) == 0) {
+                success = 1;
+#ifdef DEBUG
+                bftpd_log("Extended passive mode: Successfully bound port %d\n", port);
+#endif
+                break;
+            }
+        }   /* end of for loop */
+        if (!success) {
+            control_printf(SL_FAILURE, "425 Error: Unable to bind data socket.");
+            return;
+        }
+    }    /* end of else using list of ports */
+
+	
+        prepare_sock(pasvsock);
 	if (listen(pasvsock, 1)) {
 		control_printf(SL_FAILURE, "500-Error: Unable to make socket listen.\r\n425 %s",
 				 strerror(errno));
@@ -480,7 +527,7 @@ void command_mput(char *filenames)
    char filename[MAXCMD];  /* single filename */
    int from_index, to_index;      /* position in "filenames" and "filename" */
                                                                                                                              
-   from_index = 0;     /* start at begining of filenames */
+   from_index = 0;     /* start at beginning of filenames */
    memset(filename, 0, MAXCMD);       /* clear filename */
    to_index = 0;
 
@@ -532,15 +579,15 @@ void command_mput(char *filenames)
 void do_stor(char *filename, int flags)
 {
     char *buffer;
-    int fd, i, max;
+    int fd = -1, i, max;
     fd_set rfds;
     struct timeval tv;
     char *p, *pp;
     char *mapped = bftpd_cwd_mappath(filename);
 
-    int my_buffer_size;    /* total transfer buffer size divided by number of clients */
+    int my_buffer_size;  /* total transfer buffer size divided by number of clients */
     int num_clients = 1;       /* number of clients connected to the server */
-    int new_num_clients = 1;
+    int new_num_clients;
     int xfer_delay;
     int attempt_gzip = FALSE;
     unsigned long get_value;
@@ -551,6 +598,12 @@ void do_stor(char *filename, int flags)
     gzFile my_zip_file = NULL;
     #endif
 
+   
+    if (! mapped)
+    {
+       control_printf(SL_FAILURE, "451 Error: Unable to locate file.");
+       return;
+    } 
     if (pre_write_script)
        run_script(pre_write_script, mapped);
 
@@ -626,7 +679,7 @@ void do_stor(char *filename, int flags)
            }
            if (! my_zip_file)
            {
-              control_printf(SL_FAILURE, "553 Error: An error occured creating compressed file.");
+              control_printf(SL_FAILURE, "553 Error: An error occurred creating compressed file.");
               close(sock);
               close(fd);
               return;
@@ -668,10 +721,12 @@ void do_stor(char *filename, int flags)
     if (! buffer)
     {
        bftpd_log("Unable to create buffer to receive file.\n", 0);
-       control_printf(SL_FAILURE, "553 Error: An unknown error occured on the server.");
+       control_printf(SL_FAILURE, "553 Error: An unknown error occurred on the server.");
        if (fd >= 0)
           close(fd);
        close(sock);
+       if (mapped)
+          free(mapped);
        return;
     }
 
@@ -768,10 +823,11 @@ void do_stor(char *filename, int flags)
 	free(buffer);
         #ifdef HAVE_ZLIB_H
           gzclose(my_zip_file);
-        #else
-	  close(fd);
         #endif
 
+        if (fd >= 0)
+          close(fd);
+
 	close(sock);
         alarm(control_timeout);
         offset = 0;
@@ -854,7 +910,7 @@ void command_mget(char *filenames)
 void command_retr(char *filename)
 {
         int num_clients = 1;
-        int new_num_clients = 1;   /* number of connectiosn to the server */
+        int new_num_clients;      /* number of connections to the server */
         int my_buffer_size;       /* size of the transfer buffer to use */
 	char *mapped = NULL;
 	char *buffer;
@@ -868,7 +924,8 @@ void command_retr(char *filename)
     gzFile gzfile;
 #endif
 	int phile;
-	int i, whattodo = DO_NORMAL;
+	ssize_t i;
+        int whattodo = DO_NORMAL;
 	struct stat statbuf;
 #if (defined(WANT_TAR) && defined(WANT_GZIP))
     int filedes[2];
@@ -893,7 +950,7 @@ void command_retr(char *filename)
         if (! mapped)
         {
            bftpd_log("Memory error in sending file.\n", 0);
-           control_printf(SL_FAILURE, "553 An unknown error occured on the server.", 9);
+           control_printf(SL_FAILURE, "553 An unknown error occurred on the server.", 9);
            return;
         }
 
@@ -988,7 +1045,7 @@ void command_retr(char *filename)
                    if (mapped)
                       free(mapped);
                    bftpd_log("Memory error in sending file.\n", 0);
-                   control_printf(SL_FAILURE, "553 An unknown error occured on the server.", 9);
+                   control_printf(SL_FAILURE, "553 An unknown error occurred on the server.", 9);
                    return;
                 }
 
@@ -1003,7 +1060,8 @@ void command_retr(char *filename)
                 
                 close(filedes[1]);
                 gzfile = gzdopen(sock, "wb");
-                while ((i = read(filedes[0], buffer, my_buffer_size))) {
+                i = read(filedes[0], buffer, my_buffer_size);
+                while (i > 0) {
                     gzwrite(gzfile, buffer, i);
                     test_abort(1, phile, sock);
 
@@ -1024,6 +1082,7 @@ void command_retr(char *filename)
                         wait_time.tv_usec = xfer_delay;
                         select( 0, NULL, NULL, NULL, &wait_time);
                     }
+                    i = read(filedes[0], buffer, my_buffer_size);
                 }     // end of while 
                 free(buffer);
                 gzclose(gzfile);
@@ -1085,7 +1144,7 @@ void command_retr(char *filename)
             if (! buffer)
             {
                bftpd_log("Memory error while sending file.", 0);
-               control_printf(SL_FAILURE, "553 An unknown error occured on the server.", 0);
+               control_printf(SL_FAILURE, "553 An unknown error occurred on the server.", 0);
                if (phile) close(phile);
                return;
             }
@@ -1101,7 +1160,8 @@ void command_retr(char *filename)
 
             /* Use "wb9" for maximum compression, uses more CPU time... */
             gzfile = gzdopen(sock, "wb");
-            while ((i = read(phile, buffer, my_buffer_size))) {
+            i = read(phile, buffer, my_buffer_size);
+            while (i > 0) {
                 gzwrite(gzfile, buffer, i);
                 test_abort(1, phile, sock);
                 if (change_buffer_size)
@@ -1120,6 +1180,7 @@ void command_retr(char *filename)
                     wait_time.tv_usec = xfer_delay;
                     select( 0, NULL, NULL, NULL, &wait_time);
                 }
+                i = read(phile, buffer, my_buffer_size);
             }
             free(buffer);
             close(phile);
@@ -1137,7 +1198,7 @@ void command_retr(char *filename)
                {
                    close(phile);
                    bftpd_log("Memory error while sending file.", 0);
-                   control_printf(SL_FAILURE, "553 An unknown error occured on the server.", 0);
+                   control_printf(SL_FAILURE, "553 An unknown error occurred on the server.", 0);
                    return;
                }
 
@@ -1148,7 +1209,7 @@ void command_retr(char *filename)
                   close(phile);
                   gzclose(gzfile);
                   bftpd_log("Memory error while sending file.", 0);
-                  control_printf(SL_FAILURE, "553 An unknown error occured on the server.", 0);
+                  control_printf(SL_FAILURE, "553 An unknown error occurred on the server.", 0);
                   return;
                }
 
@@ -1210,7 +1271,7 @@ void command_retr(char *filename)
                         /* make sure buffer was created */
                         if (! buffer)
                         {
-                            control_printf(SL_FAILURE, "553 An unknown error occured.");
+                            control_printf(SL_FAILURE, "553 An unknown error occurred.");
                             bftpd_log("Memory error while trying to send file.", 0);
                             close(sock);
                             close(phile);
@@ -1224,16 +1285,17 @@ void command_retr(char *filename)
                         else
                             my_buffer_size = xfer_bufsize;
 
-			while ((i = read(phile, buffer, my_buffer_size))) {
+                        i = read(phile, buffer, my_buffer_size);
+			while (i > 0) {
 				if (test_abort(1, phile, sock)) {
 					free(buffer);
 					return;
 				}
 
-                if (xfertype == TYPE_ASCII) {
-                    buffer[i] = '\0';
-                    i += replace(buffer, "\n", "\r\n", xfer_bufsize);
-                }
+                               if (xfertype == TYPE_ASCII) {
+                                   buffer[i] = '\0';
+                                   i += replace(buffer, "\n", "\r\n", xfer_bufsize);
+                               }
 				send_status = send(sock, buffer, i, 0);
                                 // check for dropped connection
                                 if (send_status < 0)
@@ -1263,7 +1325,7 @@ void command_retr(char *filename)
                                    wait_time.tv_usec = xfer_delay;
                                    select( 0, NULL, NULL, NULL, &wait_time);
                                 }
-
+                                i = read(phile, buffer, my_buffer_size);
 			}       // end of while
             free(buffer);
             }
@@ -1309,6 +1371,12 @@ void do_dirlist(char *dirname, char verbose)
 		dirlist("*", datastream, verbose, show_hidden);
 	else {
 		char *mapped = bftpd_cwd_mappath(dirname);
+                if (! mapped)
+                {
+                   control_printf(SL_FAILURE, "451 Error: Unable to locate file.");
+                   fclose(datastream);
+                   return;
+                }
 		dirlist(mapped, datastream, verbose, show_hidden);
 		free(mapped);
 	}
@@ -1337,6 +1405,11 @@ void command_mdtm(char *filename)
 	struct stat statbuf;
 	struct tm *filetime;
 	char *fullfilename = bftpd_cwd_mappath(filename);
+        if (! fullfilename)
+        {
+           control_printf(SL_FAILURE, "451 Error: Unable to locate file.");
+           return;
+        }
 	if (!stat(fullfilename, (struct stat *) &statbuf)) {
 		filetime = gmtime((time_t *) & statbuf.st_mtime);
 		control_printf(SL_SUCCESS, "213 %04i%02i%02i%02i%02i%02i",
@@ -1373,6 +1446,11 @@ void command_dele(char *filename)
 {
         struct stat sbuf;
 	char *mapped = bftpd_cwd_mappath(filename);
+        if (! mapped)
+        {
+             control_printf(SL_FAILURE, "451 Error: Unable to perform delete..");
+             return;
+        }
         if (pre_write_script)
            run_script(pre_write_script, mapped);
 
@@ -1417,7 +1495,11 @@ void command_dele(char *filename)
 void command_mkd(char *dirname)
 {
 	char *mapped = bftpd_cwd_mappath(dirname);
-
+        if (! mapped)
+        {
+           control_printf(SL_FAILURE, "451 Error: Unable to create directory.");
+           return;
+        }
         if (pre_write_script)
            run_script(pre_write_script, mapped);
 
@@ -1439,6 +1521,12 @@ void command_mkd(char *dirname)
 void command_rmd(char *dirname)
 {
 	char *mapped = bftpd_cwd_mappath(dirname);
+        if (! mapped)
+        {
+           control_printf(SL_FAILURE, "451 Error: Unable to remove directory.");
+           return;
+        }
+
         if (pre_write_script)
            run_script(pre_write_script, mapped);
 
@@ -1477,6 +1565,12 @@ void command_rnfr(char *oldname)
 {
 	FILE *file;
 	char *mapped = bftpd_cwd_mappath(oldname);
+        if (! mapped)
+        {
+           control_printf(SL_FAILURE, "451 Error: Unable to locate file.");
+           return;
+        }
+
 	if ((file = fopen(mapped, "r"))) {
 		fclose(file);
 		if (philename)
@@ -1493,16 +1587,24 @@ void command_rnfr(char *oldname)
 void command_rnto(char *newname)
 {
 	char *mapped = bftpd_cwd_mappath(newname);
+        if ( (! mapped) || (! philename) )
+        {
+           if (mapped) free(mapped);   // avoid leaking memory
+           control_printf(SL_FAILURE, "451 Error: Unable to rename file.");
+           return;
+        }
 
         if (pre_write_script)
            run_script(pre_write_script, mapped);
 
 	if (rename(philename, mapped)) {
 		bftpd_log("Error: '%s' while trying to rename '%s' to '%s'.\n",
-				  strerror(errno), philename, bftpd_cwd_mappath(newname));
+                strerror(errno), philename, mapped);
+				  // strerror(errno), philename, bftpd_cwd_mappath(newname));
 		control_printf(SL_FAILURE, "451 Error: %s.", strerror(errno));
 	} else {
-		bftpd_log("Successfully renamed '%s' to '%s'.\n", philename, bftpd_cwd_mappath(newname));
+		bftpd_log("Successfully renamed '%s' to '%s'.\n", // philename, bftpd_cwd_mappath(newname));
+                philename, mapped);
 		control_printf(SL_SUCCESS, "250 OK");
 		state = STATE_AUTHENTICATED;
 	}
@@ -1525,8 +1627,14 @@ void command_size(char *filename)
 {
 	struct stat statbuf;
 	char *mapped = bftpd_cwd_mappath(filename);
+        if (! mapped)
+        {
+           control_printf(SL_FAILURE, "451 Error: Unable to locate file.");
+           return;
+        }
+
 	if (!stat(mapped, &statbuf)) {
-		control_printf(SL_SUCCESS, "213 %i", (int) statbuf.st_size);
+		control_printf(SL_SUCCESS, "213 %lld", (long long) statbuf.st_size);
 	} else {
 		control_printf(SL_FAILURE, "550 Error: %s.", strerror(errno));
 	}
@@ -1545,6 +1653,12 @@ void command_quit(char *params)
 void command_stat(char *filename)
 {
 	char *mapped = bftpd_cwd_mappath(filename);
+        if (! mapped)
+        {
+           control_printf(SL_FAILURE, "451 Error: Unable to discover status.");
+           return;
+        }
+
     control_printf(SL_SUCCESS, "213-Status of %s:", filename);
     bftpd_stat(mapped, stderr);
     control_printf(SL_SUCCESS, "213 End of Status.");
@@ -1566,12 +1680,17 @@ void command_chmod(char *params)
         my_string = strdup(strchr(params, ' ') + 1);
         if (! my_string)
         {
-            control_printf(SL_FAILURE, "550: An error occured on the server trying to CHMOD.");
+            control_printf(SL_FAILURE, "550: An error occurred on the server trying to CHMOD.");
             return;
         }
 	/* mapped = bftpd_cwd_mappath(strdup(strchr(params, ' ') + 1)); */
         mapped = bftpd_cwd_mappath(my_string);
         free(my_string);
+        if (! mapped)
+        {
+           control_printf(SL_FAILURE, "451 Error: Unable to locate file.");
+           return;
+        }
        
         if (pre_write_script)
            run_script(pre_write_script, mapped);
@@ -1618,6 +1737,11 @@ void command_chown(char *params)
 			return;
 		}
 	mapped = bftpd_cwd_mappath(filename);
+        if (! mapped)
+        {
+           control_printf(SL_FAILURE, "451 Error: Unable to locate file.");
+           return;
+        }
         if (pre_write_script)
            run_script(pre_write_script, mapped);
 
@@ -1706,7 +1830,7 @@ void command_site(char *str)
             control_printf(SL_SUCCESS, "211-Possible usages for SITE command:");
             control_printf(SL_SUCCESS, "211-site chmod <mask> <filename>");
             control_printf(SL_SUCCESS, "211-site chown <owner> <filename>");
-            control_printf(SL_SUCCESS, "211 site md5 <filename>");
+            control_printf(SL_SUCCESS, "211-site md5 <filename>");
             return;
         }
 	control_printf(SL_FAILURE, "550 Unknown command: 'SITE %s'.", str);
@@ -1867,7 +1991,7 @@ int get_buffer_size(int num_connections)
 
 /*
 This function forks and runs a script. On success it
-returns TRUE, if an error occures, it returns FALSE.
+returns TRUE, if an error occurs, it returns FALSE.
 */
 int run_script(char *script, char *path)
 {
@@ -1906,7 +2030,8 @@ int run_script(char *script, char *path)
    /* parent process */
    do
    {
-       process_id = wait4(process_id, NULL, 0, NULL);
+       // process_id = wait4(process_id, NULL, 0, NULL);
+       process_id = waitpid(process_id, NULL, 0);
    } while ( (process_id == -1) && (errno == EINTR) );
 
    signal(SIGQUIT, save_quit);
